﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using SimpleJSON;
using MVR.FileManagementSecure;
using UnityEngine.UI;


namespace JayJayWon
{
    public class FolderBrowserUI
    {
        private GameObject folderBrowserBoxGO;
        private List<FolderRowsUI> folderBrowserRows = new List<FolderRowsUI>();
        private GameObject folderListContentGO;

        private int currentResourceType;
        private int folderBrowserType;

        private BrowserFiltersUI browserFiltersUI;

        private UIDynamicButton closeButton;

        public delegate void FolderSelectedCallback(Folder folder);

        private FolderSelectedCallback folderSelectedCallback;

        private GameObject folderOptionsGO;
        private FolderRowsUI folderRowOptions;

        public FolderBrowserUI(GameObject folderBrowserBoxGO, int folderBrowserType, FolderSelectedCallback callback, BrowserFiltersUI browserFiltersUI)
        {
            this.folderBrowserBoxGO = folderBrowserBoxGO;
            this.folderBrowserType = folderBrowserType;
            folderSelectedCallback = callback;
            folderListContentGO = folderBrowserBoxGO.transform.Find("scrollableContentViewportGO/vScrollableContentGO").gameObject;
            this.browserFiltersUI = browserFiltersUI;
        }

        public void RefreshUI(int resourceType)
        {
            ClearUI();

            currentResourceType = resourceType;
            foreach (Folder folder in FolderBrowser.GetFolderWithVisibleSubFolders(resourceType))
            {

                GameObject rowGO = UI.AddChildGOImage(folderListContentGO, new Color(0f, 0.8f, 0f, 0f), AnchorPresets.hStretchTop, 500f, 20f, Vector2.zero);
                folderBrowserRows.Add(new FolderRowsUI(rowGO, folder, this, browserFiltersUI, folderSelectedCallback));
            }
            closeButton = UI.CreateUIButton(folderBrowserBoxGO, 20f, 20f, "X", 28, 0, -20f, AnchorPresets.bottomRight);
            browserFiltersUI.parentBrowser.ApplyHelp(closeButton, "Close the Folder Browser window");
            closeButton.button.onClick.AddListener(delegate
            {
                folderBrowserBoxGO.SetActive(false);
            });
        }
        public void ClearUI()
        {
            foreach (FolderRowsUI folderRow in folderBrowserRows)
            {
                folderRow.ClearUI();
            }
            folderBrowserRows.Clear();
            if (closeButton != null) GameObject.Destroy(closeButton.gameObject);
            folderRowOptions = null;
            if (folderOptionsGO != null) GameObject.Destroy(folderOptionsGO);
        }

        public void RefreshUICurrentResourceType()
        {
            RefreshUI(currentResourceType);
        }

        

        public void ToggleOptionsMenu(FolderRowsUI folderRow)
        {

            if (folderRow == folderRowOptions)
            {
                folderRowOptions = null;
                if (folderOptionsGO != null) GameObject.Destroy(folderOptionsGO);
            }
            else
            {
                folderRowOptions = null;
                if (folderOptionsGO != null) GameObject.Destroy(folderOptionsGO);

                folderOptionsGO = UI.AddChildGOImage(folderBrowserBoxGO, Color.white, AnchorPresets.topLeft, 103f, (folderRow.optionsMenuItems*20f)+1.5f, new Vector2(-100f,-20f));
                folderOptionsGO.transform.position = folderRow.optionsButton.transform.position + (folderRow.optionsButton.transform.up *-10f) + folderRow.optionsButton.transform.right * -110f;

                float offset = 1.5f;
                if (folderRow.isRemoveRenameFolderOption)
                {
                    var renameFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Rename Folder...",20, 0, -offset, anchorPreset:AnchorPresets.topMiddle,clearSprite:true);
                    renameFolderButton.button.onClick.AddListener(delegate
                    {
                        folderRow.RenameFolderStart();
                        
                    });
                    var deleteFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Delete Folder...",20,0, -offset - 20f,AnchorPresets.topMiddle, clearSprite: true);
                    deleteFolderButton.buttonColor = Color.red;
                    deleteFolderButton.textColor = Color.white;
                    var moveFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Move Folder...", 20, 0, -offset - 40f, AnchorPresets.topMiddle, clearSprite: true);
                    offset = 60f;
                }
                if (folderRow.isCreateFolderOption)
                {
                    var createFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Create Sub-folder", 20, 0, -offset, AnchorPresets.topMiddle, clearSprite: true);
                    createFolderButton.button.onClick.AddListener(delegate
                    {
                        folderRow.CreateSubFolder();
                        RefreshUICurrentResourceType();
                    });
                    offset += 20f;
                }
                if (folderRow.isExpandCollapseFolderOption)
                {
                    var expandFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Expand all Sub-folders", 20, 0, -offset, AnchorPresets.topMiddle, clearSprite: true);
                    expandFolderButton.button.onClick.AddListener(delegate
                    {
                        folderRow.ExpandAll();
                        RefreshUICurrentResourceType();
                    });
                    var collapseFolderButton = UI.CreateUIButton(folderOptionsGO, 100f, 20f, "Collapse all Sub-folders", 20, 0, -offset-20f, AnchorPresets.topMiddle, clearSprite: true);
                    collapseFolderButton.button.onClick.AddListener(delegate
                    {
                        folderRow.CollapseAll();
                        RefreshUICurrentResourceType();
                    });

                }

                    folderRowOptions = folderRow;
            }
        }
    }



    public class FolderRowsUI
    {
        private GameObject folderRowGO;
        private Folder folder;
        private FolderBrowserUI parentFolderBrowserUI;

        public UIDynamicButton optionsButton { get; private set; }

        private UIDynamicButton folderButton;

        private UIDynamicTextField folderNameTextField;

        private JSONStorableString folderRenameJSS;

        private bool isRenameActive = false;

        public FolderRowsUI(GameObject folderRowGO, Folder folder, FolderBrowserUI parentFolderBrowserUI, BrowserFiltersUI browserFiltersUI, FolderBrowserUI.FolderSelectedCallback callback)
        {
            this.folderRowGO = folderRowGO;
            this.folder = folder;
            this.parentFolderBrowserUI = parentFolderBrowserUI;

            int subFolderCount = folder.subFolderCount;
            string expandedSymbol = "+";
            if (folder.expanded) expandedSymbol = "-";

            int depth = folder.GetDepth();

            folderRenameJSS = new JSONStorableString("folderRename", "", FolderRenameUpdate);

            if (subFolderCount > 0)
            {
                var expandButton = UI.CreateUIButton(folderRowGO, 20f, 20f, expandedSymbol, 28, depth * 10f, 0, AnchorPresets.middleLeft);
                browserFiltersUI.parentBrowser.ApplyHelp(expandButton, "Expand or collapse sub folders in the tree view for this folder");

                expandButton.button.onClick.AddListener(delegate
                {
                    folder.ToggleExpanded();
                    parentFolderBrowserUI.RefreshUICurrentResourceType();
                });
            }

            float buttonWidth = 150f;
            if (folder.isRootFolder) buttonWidth = 250f;

            folderButton = UI.CreateUIButton(folderRowGO, buttonWidth, 20f, "  " + folder.folderName, 20, (depth * 10f) + 20f, 0, AnchorPresets.middleLeft, false);

            if (browserFiltersUI.selectedFolderFilter == folder)
            {
                folderButton.buttonText.text = "<b>  " + folder.folderName + "</b>";
                folderButton.buttonColor = Color.cyan;
            }
            browserFiltersUI.parentBrowser.ApplyHelp(folderButton, "Select " + folder.GetFolderPath());
            folderButton.buttonText.alignment = TextAnchor.MiddleLeft;
            folderButton.button.onClick.AddListener(delegate
            {
                callback(folder);
            });

            

            folderNameTextField = UI.CreateTextBox(folderRowGO, folderRenameJSS, buttonWidth, 20f, 20, (depth * 10f) + 20f, 0, AnchorPresets.middleLeft, true, InputField.LineType.SingleLine,"",0f,0f);
            folderNameTextField.gameObject.SetActive(false);

/*            var inputComponent = folderNameTextField.gameObject.GetComponent<InputField>();
            var cancelFolderNameButton = UI.CreateUIButton(folderNameTextField.gameObject, 20f, 20f, "X", 20, 0, 0, AnchorPresets.middleRight);
            cancelFolderNameButton.button.onClick.AddListener(delegate
            {
                SuperController.LogMessage("Cancel button");
            });

            */


            if (optionsMenuItems > 0)
            {
                optionsButton = UI.CreateUIButton(folderRowGO, 20f, 20f, "v", 20, (depth * 10f) + 20f + buttonWidth, 0, AnchorPresets.middleLeft, false);
                optionsButton.buttonText.alignment = TextAnchor.MiddleCenter;
                optionsButton.button.onClick.AddListener(delegate
                {
                    if (!isRenameActive) parentFolderBrowserUI.ToggleOptionsMenu(this);
                    else RenameFolderEnd();
                });
            }
            


        }

        private string updatedFolderName;
        private void FolderRenameUpdate(string newFolderName)
        {
            SuperController.LogMessage("FolderRenameUpdate");
            string folderPath = folder.GetFolderPath();

            updatedFolderName = newFolderName;
            FileManagerSecure.MoveFile(Const.SymLinksFolder + "\\" + folderPath, FileManagerSecure.GetDirectoryName(Const.SymLinksFolder + "\\" + folderPath) + "\\" + newFolderName,FolderRenameConfirmed,null,null);
            
        }

        private void FolderRenameConfirmed()
        {
            SuperController.LogMessage("FolderRenameConfirmed");
            folder.RenameFolder(updatedFolderName);
            folderButton.label = "  " + folder.folderName;


            RenameFolderEnd();
        }

        public void RenameFolderEnd()
        {
            isRenameActive = false;
            folderButton.gameObject.SetActive(true);
            optionsButton.gameObject.SetActive(true);
            folderRenameJSS.valNoCallback = folder.folderName;
            folderNameTextField.gameObject.SetActive(false);
            optionsButton.buttonText.text = "v";
        }
        public void RenameFolderStart()
        {
            isRenameActive = true;
            folderNameTextField.gameObject.SetActive(true);
            folderRenameJSS.valNoCallback = folder.folderName;
            folderButton.gameObject.SetActive(false);
            optionsButton.gameObject.SetActive(false);
            
            folderNameTextField.backgroundColor = Color.yellow;

            var inputComponent = folderNameTextField.gameObject.GetComponent<InputField>();
            inputComponent.ActivateInputField();


            optionsButton.buttonText.text = "X";
            parentFolderBrowserUI.ToggleOptionsMenu(this);
        }
        public void CreateSubFolder()
        {
            folder.CreateSubFolder();
        }

        public void ExpandAll()
        {
            folder.ExpandFolderPlusChildren();
        }
        public void CollapseAll()
        {
            folder.CollapseFolderPlusChildren();
        }

        public int optionsMenuItems { get
            {
                int count = 0;
                if (isRemoveRenameFolderOption) count += 3;
                if (isCreateFolderOption) count += 1;
                if (isExpandCollapseFolderOption) count += 2;
                return count;
            } }

        private bool isWriteAccesible { get
            {
                if (folder.GetFolderPath().StartsWith("Saves\\") && !Globals.savesFolderAccess) return false;
                if (folder.GetFolderPath().StartsWith("Custom\\") && !Globals.customFolderAccess) return false;
                
                return true;
            } }
        public bool isRemoveRenameFolderOption
        {
            get {
                if (folder.isRootFolder) return false;
                if (!isWriteAccesible) return false;
                return true;
            }
        }

        public bool isCreateFolderOption
        {
            get
            {
                if (!isWriteAccesible) return false;
                return true;
            }
        }

        public bool isExpandCollapseFolderOption
        {
            get
            {
                if (folder.subFolderCount==0) return false;
                return true;
            }
        }


        public void ClearUI()
        {
            GameObject.Destroy(folderRowGO);
        }
    }

    public class FolderBrowser
    {        
        private static Dictionary<int, HashSet<Folder>> rootFoldersByResource = new Dictionary<int, HashSet<Folder>>();

        private static Dictionary<string, Folder> folders = new Dictionary<string, Folder>();

        public static HashSet<Folder> GetFolderWithVisibleSubFolders(int resourceType)
        {
            HashSet<Folder> visibleFolders = new HashSet<Folder>();

            if (!rootFoldersByResource.ContainsKey(resourceType))
            {
                CreateRootFolders(resourceType);
            }

            foreach (Folder rootFolder in rootFoldersByResource[resourceType])
            {
                visibleFolders.UnionWith(rootFolder.GetFolderWithVisibleSubFolders());
            }

            return visibleFolders;
        }

        private static void CreateRootFolders(int resourceType)
        {
            rootFoldersByResource.Add(resourceType, new HashSet<Folder>());
            foreach (string rootFolderPath in Const.GetResourceRootFolderNames(resourceType))
            {
                Folder newFolder = new Folder(rootFolderPath);
                rootFoldersByResource[resourceType].Add(newFolder);
                folders.Add(newFolder.GetFolderPath(), newFolder);
            }
        }

        private static void RescanAllRootSubFolders(int resourceType)
        {
            foreach(Folder rootFolder in rootFoldersByResource[resourceType])
            {
                rootFolder.RescanSubFolders();
            }
        }

        public static Folder GetFolder(string folderPath, int resourceType)
        {
            if (!rootFoldersByResource.ContainsKey(resourceType))
            {
                CreateRootFolders(resourceType);
            }
            if (folders.ContainsKey(folderPath)) return folders[folderPath];
            else
            {
                RescanAllRootSubFolders(resourceType);
                if (folders.ContainsKey(folderPath)) return folders[folderPath];

                if (FileManagerSecure.DirectoryExists(folderPath))
                {
                    SuperController.LogError("FolderBrowser.GetFolder: failed to find existing resource folder "+ folderPath);
                    return null;
                }
                else return rootFoldersByResource[resourceType].First();
            }
        }
        public static void RemoveFolder(Folder removeFolder)
        {
            folders.Remove(removeFolder.GetFolderPath());

        }
        public static Folder GetSubFolder(string subFolderName, Folder parentFolder)
        {
            string subFolderPath = parentFolder.GetFolderPath() + "\\" + subFolderName;
            if (folders.Keys.Contains(subFolderPath))
            {
                return folders[subFolderPath];
            }
            var newFolder = new Folder(parentFolder, subFolderName);
            folders.Add(subFolderPath, newFolder);
            return newFolder;
        }
    }

    public class Folder
    {
        public bool isRootFolder { get; private set; } = false;
        public bool expanded { get; private set; } = false;

        private Folder parentFolder = null;

        public int subFolderCount { get { return subFolders.Keys.Count; } }

        public string folderName { get; private set; }

        private Dictionary<string, Folder> subFolders = new Dictionary<string, Folder>();

        private HashSet<ResourceVersionGroupEntry> localRVGEs = new HashSet<ResourceVersionGroupEntry>();

        public Folder(string rootFolderPath)
        {
            parentFolder = null;
            folderName = rootFolderPath;
            isRootFolder = true;
            expanded = true;
            RefreshSubFolders();
        }

        public void RegisterLocalRVGE(ResourceVersionGroupEntry rvge)
        {
            if (rvge.localFileResource)
            {
                localRVGEs.Add(rvge);
            }
            else SuperController.LogError("BA.Folder.RegisterLocalRVGE: Attempt to register non local RVGE with Folder");
        }

        public void UnRegisterLocalRVGE(ResourceVersionGroupEntry rvge)
        {
            localRVGEs.Remove(rvge);
        }

        public Folder(Folder parentFolder, string folderName)
        {
            this.parentFolder = parentFolder;
            this.folderName = folderName;

            RefreshSubFolders();
        }
        public HashSet<Folder> allContainingLocalFolders
        {
            get
            {
                HashSet<Folder> result = new HashSet<Folder>();

                result.Add(this);
                if (!isRootFolder) result.UnionWith(parentFolder.allContainingLocalFolders);
                return result;
            }
        }
        public int GetDepth()
        {
            if (isRootFolder) return 0;
            return parentFolder.GetDepth() + 1;
        }

        public string GetFolderPath()
        {
            if (isRootFolder) return folderName;
            return parentFolder.GetFolderPath() + "\\" + folderName;
        }

        public HashSet<Folder> GetFolderWithVisibleSubFolders()
        {
            var visibleFolders = new HashSet<Folder>();
            visibleFolders.Add(this);
            if (expanded)
            {
                foreach (var subFolder in subFolders)
                {
                    visibleFolders.UnionWith(subFolder.Value.GetFolderWithVisibleSubFolders()); ;
                }

            }

            return visibleFolders;
        }

        public void RefreshSubFolders()
        {
            var subFolderPaths = FileManagerSecure.GetDirectories(GetFolderPath());

            var newSubFolders = new Dictionary<string, Folder>();

            foreach (var subFolderPath in subFolderPaths)
            {
                string subFolderName = subFolderPath.Substring
(subFolderPath.LastIndexOf('\\') + 1);

                if (subFolders.Keys.Contains(subFolderName)) newSubFolders.Add(subFolderName, subFolders[subFolderName]);
                else newSubFolders.Add(subFolderName, FolderBrowser.GetSubFolder(subFolderName, this));
            }
            var missingFolders = subFolders.Values.Except(newSubFolders.Values) as HashSet<Folder>;

            if (missingFolders != null) RemoveSubFolders(missingFolders);
            subFolders = newSubFolders;
        }
        public void RescanSubFolders()
        {
            RefreshSubFolders();
            foreach (var subFolder in subFolders)
            {
                subFolder.Value.RescanSubFolders();
            }
        }

        public void RenameFolder(string newFolderName)
        {
            folderName = newFolderName;
            UpdateRVGEFolderPaths();
        }

        private void UpdateRVGEFolderPaths()
        {
            foreach(LocalRVGE rvge in localRVGEs)
            {
                rvge.ResetFolderPath();
            }
            foreach(var subFolderKVP in subFolders)
            {
                subFolderKVP.Value.UpdateRVGEFolderPaths();
            }
        }

        public void RemoveFolder()
        {
            var allSubFolders = new HashSet<Folder>(subFolders.Values.ToArray());
            RemoveSubFolders(allSubFolders);
            FolderBrowser.RemoveFolder(this);
        }
        private void RemoveSubFolders(HashSet<Folder> removeSubFolders)
        {
            foreach (Folder subFolder in removeSubFolders)
            {
                subFolder.RemoveFolder();
            }
        }

        public void ToggleExpanded()
        {
            if (!expanded)
            {
                RefreshSubFolders();
            }
            expanded = !expanded;
        }

        public void ExpandParents()
        {

            if (!isRootFolder) parentFolder.ExpandFolderPlusParents();
        }

        public void ExpandFolderPlusParents()
        {
            expanded = true;
            if (!isRootFolder) parentFolder.ExpandFolderPlusParents();
        }

        public void ExpandFolderPlusChildren()
        {
            expanded = true;
            foreach (var subFolder in subFolders) subFolder.Value.ExpandFolderPlusChildren();
        }

        public void CollapseFolderPlusChildren()
        {
            expanded = false;
            foreach (var subFolder in subFolders) subFolder.Value.CollapseFolderPlusChildren();
        }
        public void CreateSubFolder()
        {
            expanded = true;
            string folderName = "New Folder";
            int folderNumber = 1;
            while (FileManagerSecure.DirectoryExists(Const.SymLinksFolder + "\\" + GetFolderPath() + "\\" + folderName))
            {
                folderName = "New Folder(" + folderNumber.ToString() + ")";
                folderNumber++;
            }
            FileManagerSecure.CreateDirectory(Const.SymLinksFolder + "\\" + GetFolderPath() + "\\" + folderName);
            RefreshSubFolders();
        }
    }
}
